#ifndef _H_AUDIO_CHANNELS
#define _H_AUDIO_CHANNELS

#include <celt051/celt.h>

#include "redchannel.h"
#include "debug.h"

class ChannelFactory;

class WavePlaybackAbstract;
class WaveRecordAbstract;
class RecordSamplesMessage;

class PlaybackChannel: public RedChannel {
public:
    PlaybackChannel(RedClient& client, uint32_t id);
    ~PlaybackChannel(void);
    bool abort(void);

    static ChannelFactory& Factory();

private:
    void handle_mode(RedPeer::InMessage* message);
    void handle_start(RedPeer::InMessage* message);
    void handle_stop(RedPeer::InMessage* message);
    void handle_raw_data(RedPeer::InMessage* message);
    void handle_celt_data(RedPeer::InMessage* message);
    void null_handler(RedPeer::InMessage* message);
    void disable();

    void set_data_handler();

private:
    WavePlaybackAbstract* _wave_player;
    uint32_t _mode;
    uint32_t _frame_bytes;
    CELTMode *_celt_mode;
    CELTDecoder *_celt_decoder;
    bool _playing;
    uint32_t _frame_count;
};

class RecordChannel : public RedChannel, private Platform::RecordClinet {
public:
    RecordChannel(RedClient& client, uint32_t id);
    ~RecordChannel(void);
    
    bool abort(void);

    static ChannelFactory& Factory();

private:
    void handle_start(RedPeer::InMessage* message);
    void handle_stop(RedPeer::InMessage* message);
 
    virtual void on_connect();

    virtual void add_evnet_sorce(EventsLoop::File& evnet_sorce);
    virtual void remove_evnet_sorce(EventsLoop::File& evnet_sorce);
    virtual void add_evnet_sorce(EventsLoop::Trigger& evnet_sorce);
    virtual void remove_evnet_sorce(EventsLoop::Trigger& evnet_sorce);
    virtual void push_frame(uint8_t *frame);

    void send_start_mark();
    void release_message(RecordSamplesMessage *message);
    RecordSamplesMessage * get_message();

private:
    WaveRecordAbstract* _wave_recorder;
    Mutex _messages_lock;
    std::list<RecordSamplesMessage *> _messages;
    int _mode;
    CELTMode *_celt_mode;
    CELTEncoder *_celt_encoder;
    uint32_t _frame_bytes;

    static int data_mode;

    friend class RecordSamplesMessage;
};

#endif

