#ifndef _H_CURSOR_
#define _H_CURSOR_

#include <boost/detail/atomic_count.hpp>

#include "red.h"
#include "red_window_p.h"

class CursorOpaque {
public:
    CursorOpaque() {}
    virtual ~CursorOpaque() {}
};

class LocalCursor;

class CursorData {
public:
    CursorData(RedCursor& cursor, int data_size);
    
    CursorData *ref() { ++_atomic; return this;}
    void unref() {if (--_atomic == 0) delete this;}
    void set_opaque(CursorOpaque* opaque) { delete _opaque; _opaque = opaque;}
    CursorOpaque* get_opaque() { return _opaque;}
    void set_local(LocalCursor* local_cursor);
    LocalCursor* get_local() { return _local_cursor;} 
    const CursorHeader& header() const { return _header;};
    const uint8_t* data() const { return _data;}
    
private:
    ~CursorData();
    
private:
    boost::detail::atomic_count _atomic;
    CursorHeader _header;
    uint8_t* _data;
    CursorOpaque* _opaque;
    LocalCursor* _local_cursor;
};

class LocalCursor {
public:
    LocalCursor() : _atomic(1) {}
    virtual ~LocalCursor() {}
    virtual void set(Window window) {}
    LocalCursor* ref() { ++_atomic; return this;}
    void unref() { if (--_atomic == 0) delete this;}
protected:
    static int get_size_bits(const CursorHeader& header, int &size);
private:
    boost::detail::atomic_count _atomic;
};

#endif
