#ifndef _H_GCANVAS
#define _H_GCANVAS

#include "canvas.h"
#include "cairo_canvas.h"
#include "gl_canvas.h"
#include "red_pixmap_gl.h"
#include "red_window.h"

class RedPixmapGL;

class GCanvas : public Canvas {
public:
    GCanvas(PixmapCache& pixmap_cache, PaletteCache& palette_cache,
            GlzDecoderWindow &glz_decoder_window);
    virtual ~GCanvas();

    void set_mode(int width, int height, int depth, RedWindow *win,
                  RenderType rendertype);
    void clear();
    void thread_touch() {}
    void copy_pixels(const QRegion& region, RedDrawable* dc,
                     const PixmapHeader* pixmap);
    void copy_pixels(const QRegion& region, RedDrawable& dc);
    void put_image(const PixmapHeader& image, const Rect& dest,
                   const QRegion* clip);

    void set_access_params(unsigned long base, unsigned long max);
    void draw_fill(Rect *bbox, Clip *clip, Fill *fill);
    void draw_copy(Rect *bbox, Clip *clip, Copy *copy);
    void draw_opaque(Rect *bbox, Clip *clip, Opaque *opaque);
    void copy_bits(Rect *bbox, Clip *clip, Point *src_pos);
    void draw_text(Rect *bbox, Clip *clip, Text *text);
    void draw_stroke(Rect *bbox, Clip *clip, Stroke *stroke);
    void draw_rop3(Rect *bbox, Clip *clip, Rop3 *rop3);
    void draw_blend(Rect *bbox, Clip *clip, Blend *blend);
    void draw_blackness(Rect *bbox, Clip *clip, Blackness *blackness);
    void draw_whiteness(Rect *bbox, Clip *clip, Whiteness *whiteness);
    void draw_invers(Rect *bbox, Clip *clip, Invers *invers);
    void draw_transparent(Rect *bbox, Clip *clip, Transparent* transparent);
    void draw_alpha_blend(Rect *bbox, Clip *clip, AlphaBlnd* alpha_blend);

    virtual void destroy_textures();
    virtual CanvasType get_pixmap_type();
    virtual void touch_context();
    virtual void pre_gl_copy();
    virtual void post_gl_copy();

private:
    void create_pixmap(int width, int height, RedWindow *win,
                       RenderType rendertype);
    void destroy_pixmap();
    void destroy();
   
private:
    GLCanvas* _canvas;
    RedPixmapGL *_pixmap;
};

#endif

