#ifndef _H_GLZ_DECODED_IMAGE
#define _H_GLZ_DECODED_IMAGE

#include "common.h"
#include "glz_decoder_config.h"

/*
    This class represents an image the lz window holds. It is created after the decoding of the 
    image is completed, and destroyed when it exits the window.
*/

class GlzDecodedImage {
public:
  
    GlzDecodedImage(uint64_t id, uint64_t win_head_id, uint8_t *data, int size,
                   int bytes_per_pixel)    
    : 
    _id (id), 
    _win_head_id(win_head_id), 
    _data(data),
    _bytes_per_pixel(bytes_per_pixel),
    _size(size){}

    virtual ~GlzDecodedImage(){}
    uint8_t *get_data() {return _data;}
    uint8_t *get_pixel_ref(int offset); // pallete pix_id = byte count
    uint64_t get_id() {return _id;}
    uint64_t get_window_head_id() {return _win_head_id;}
    int      get_size() {return _size;}
        
protected:
    uint64_t _id;
    uint64_t _win_head_id;
    uint8_t *_data;
    int  _bytes_per_pixel; // if image is with palette pixel=byte
    int  _size;            // number of pixels 
};

inline uint8_t* GlzDecodedImage::get_pixel_ref(int offset) 
{
    if (!_data) {
        return NULL;
    } else {
        return (_data + (offset*_bytes_per_pixel));
    }
}

#endif
