#ifndef _H_INPUTS_CHANNEL
#define _H_INPUTS_CHANNEL

#include "redchannel.h"
#include "inputs_handler.h"

class ChannelFactory;

class InputsChannel: public RedChannel, public InputsHandler {
public:
    InputsChannel(RedClient& client, uint32_t id);
    virtual ~InputsChannel();

    virtual void on_mouse_motion(int dx, int dy, int buttons_state);
    virtual void on_mouse_position(int x, int y, int buttons_state, int display_id);
    virtual void on_mouse_down(int button, int buttons_state);
    virtual void on_mouse_up(int button, int buttons_state);
    virtual void on_key_down(uint32_t scan_code);
    virtual void on_key_up(uint32_t scan_code);
    virtual void on_focus_in();

    static ChannelFactory& Factory();

protected:
    virtual void on_connect();
    virtual void on_disconnect();
    virtual void on_migrate();

private:
    void set_motion_event(RedcMouseMotion& motion_event);
    void set_position_event(RedcMousePosition& position_event);
    void set_local_modifiers();

    void handle_init(RedPeer::InMessage* message);
    void handle_modifaiers(RedPeer::InMessage* message);
    void handle_motion_ack(RedPeer::InMessage* message);

private:
    Mutex _motion_lock;
    int _mouse_buttons_state;
    int _mouse_dx;
    int _mouse_dy;
    unsigned int _mouse_x;
    unsigned int _mouse_y;
    int _display_id;
    bool _active_motion;
    int _motion_count;
    uint32_t _modifiers;
    Mutex _update_modifiers_lock;
    bool _active_modifiers_event;

    friend class MotionMessage;
    friend class PositionMessage;
    friend class KeyModifiersEvent;
};


#endif

