#include "common.h"
#include "debug.h"
#include "peer_connect_options.h"

peer_connect_options::peer_connect_options() :
    _bIsSSL(false),
    _nServerAuthFlags(0),
    _pPubKey(NULL),
    _nPubKeySize(0)
{
    memset(_szCertFile, 0, sizeof(_szCertFile));
    memset(_szKeyFile, 0, sizeof(_szKeyFile));
    memset(_szPassword, 0, sizeof(_szPassword));
    memset(_szCAFile, 0, sizeof(_szCAFile));
    memset(_szCipher, 0, sizeof(_szCipher));
    _nPort = 0;
}

peer_connect_options::peer_connect_options(const peer_connect_options& other) :
    _bIsSSL(false),
    _nServerAuthFlags(0),
    _pPubKey(NULL),
    _nPubKeySize(0)
{
    assign(other);
}

void peer_connect_options::assign(const peer_connect_options& other)
{
    SetSSLParameters(other.IsSSL(), 
                     other.GetCertFile(),
                     other.GetKeyFile(),
                     other.GetPassword(),
                     other.GetCipher(),
                     other.GetPort());

    SetSSLServerAuthParameters(other.GetServerAuthFlags(),
                               other.GetPubKey(),
                               other.GetPubKeySize(),
                               other.GetCAFile(),
                               other.GetSubject());
}

peer_connect_options::~peer_connect_options()
{
    if (_pPubKey) {
        delete []_pPubKey;
    }
}

void peer_connect_options::SetSSLParameters (
	bool bIsSSL,
	const char *szCertFile, 
	const char *szKeyFile, 
	const char *szPassword, 
	const char *szCipher,
	const uint16_t nPort
) {
	_bIsSSL = bIsSSL;

	if (szCertFile != NULL) strncpy (_szCertFile, szCertFile, sizeof (_szCertFile));
	if (szKeyFile != NULL) strncpy (_szKeyFile, szKeyFile, sizeof (_szKeyFile));
	if (szPassword != NULL) strncpy (_szPassword, szPassword, sizeof (_szPassword));
	if (szCipher != NULL) strncpy (_szCipher, szCipher, sizeof (_szCipher));

	_nPort = nPort;
}

void peer_connect_options::SetSSLServerAuthParameters (
    int nServerAuthFlags,
    const uint8_t *pPubKey,
    int nPubKeySize,
    const char *szCAFile,
    const CertFieldValueList& lSubject)
{
    ASSERT(!(nServerAuthFlags & SERVER_AUTH_PUBKEY) || (pPubKey && nPubKeySize));
    ASSERT(!(nServerAuthFlags & SERVER_AUTH_NAME) || szCAFile);
    ASSERT(!(nServerAuthFlags & SERVER_AUTH_SUBJECT) || (szCAFile && !lSubject.empty()));
    ASSERT(((nServerAuthFlags & SERVER_AUTH_NAME) || (nServerAuthFlags & SERVER_AUTH_SUBJECT)) ||
           (!szCAFile));

    _nServerAuthFlags = nServerAuthFlags;
    if (_pPubKey) {
        delete []_pPubKey;
        _pPubKey = NULL;
        _nPubKeySize = 0;
    }

    if (nPubKeySize) {
        _pPubKey = new uint8_t[nPubKeySize];
        memcpy(_pPubKey, pPubKey, nPubKeySize);
        _nPubKeySize = nPubKeySize;
    }

    if (szCAFile != NULL) {
        strncpy (_szCAFile, szCAFile, sizeof (_szCAFile));
    } else {
        _szCAFile[0] = 0;
    }

    _lSubject = lSubject;
}
